#include "../std.h"
#include <dlfcn.h>

#include "../jobs/users.h"
#include "../jobs/states.h"
#include <sys/resource.h>
#include <sys/wait.h>
#include <globus_common.h>
#include "../config/environment.h"
#include "../config/conf.h"
#include "../misc/substitute.h"
#include "../misc/log_time.h"
#include "run.h"

extern char** environ;

static int job_counter = 0;
bool RunParallel::hup_detected = false;

bool RunParallel::in_signal  = false;
RunElement* RunParallel::begin   = new RunElement;
RunElement* RunParallel::begin_u = NULL;
globus_cond_t* RunParallel::cond  = NULL;

static RunParallel run_parallel; // father of all - it must be at least one

/* Linux specific - stupid g++ limitation */
// struct sigaction RunParallel::old_sig_chld = { SIG_ERR, { }, ~SA_SIGINFO };
// struct sigaction RunParallel::old_sig_hup  = { SIG_ERR, { }, ~SA_SIGINFO };
struct sigaction RunParallel::old_sig_chld = {  };
struct sigaction RunParallel::old_sig_hup  = {  };
struct sigaction RunParallel::old_sig_term  = {  };
bool RunParallel::old_sig_chld_inited = false;
bool RunParallel::old_sig_hup_inited  = false;
bool RunParallel::old_sig_term_inited  = false;

RunElement* RunParallel::add_handled(RunElement* re) {
  if(begin_u) { begin->next_u=begin_u; };
  begin_u=begin;
  re->next=begin; begin=re;
  return begin_u;
}

void RunParallel::release(RunElement* re) {
  if(re != NULL) {
    re->released=true;
  };
  /* remove all with pid=-1 && released=true */
  for(RunElementPointer p=begin_u;!p.is_end();) {
    if((p->pid == -1) && (p->released)) {
      if(p.prev) {
        p.prev->next_u=p.cur->next_u;
        delete p.cur;
        p.cur=p.prev;
        p.go_next_u();
      }
      else { 
        begin_u=p.cur->next_u;
        delete p.cur;
        p=begin_u;
      };
    }
    else { p.go_next_u(); };
  };
}

typedef struct {
  const JobUser* user;
  const JobDescription* job;
  const char* reason;
} job_subst_t;

static void job_subst(string& str,void* arg) {
  job_subst_t* subs = (job_subst_t*)arg;
  for(int p = 0;;) {
    p=str.find('%',p);
    if(p==string::npos) break;
    if(str[p+1]=='I') {
      str.replace(p,2,subs->job->get_id().c_str());
      p+=subs->job->get_id().length();
    } else if(str[p+1]=='S') {
      str.replace(p,2,subs->job->get_state_name());
      p+=strlen(subs->job->get_state_name());
    } else if(str[p+1]=='O') {
      str.replace(p,2,subs->reason);
      p+=strlen(subs->reason);
    } else {
      p+=2;
    };
  };
  subs->user->substitute(str);
}

bool RunParallel::run(JobUser& user,const JobDescription& desc,char *const args[],RunElement** ere,bool su) {
  RunPlugin* cred = user.CredPlugin();
  job_subst_t subs; subs.user=&user; subs.job=&desc; subs.reason="external";
  if((!cred) || (!(*cred))) { cred=NULL; };
  if(user.get_uid() == 0) {
    JobUser tmp_user(desc.get_uid());
    if(!tmp_user.is_valid()) return false;
    tmp_user.SetControlDir(user.ControlDir());
    tmp_user.SetSessionRoot(user.SessionRoot());
    return run(tmp_user,desc.get_id().c_str(),args,ere,su,
                                        true,cred,&job_subst,&subs);
  };
  return run(user,desc.get_id().c_str(),args,ere,su,
                                      true,cred,&job_subst,&subs);
}

/* fork & execute child process with stderr redirected 
   to job.ID.errors, stdin and stdout to /dev/null */
bool RunParallel::run(JobUser& user,const char* jobid,char *const args[],RunElement** ere,bool su,bool job_proxy,RunPlugin* cred,RunPlugin::substitute_t subst,void* subst_arg) {
  /* create slot */
  RunElement* re = new RunElement;
  (*ere)=NULL;
  re=add_handled(re);
  pid_t* p_pid = &(re->pid);
  { sigset_t sig; sigemptyset(&sig);sigaddset(&sig,SIGCHLD); sigprocmask(SIG_BLOCK,&sig,NULL); perror("sigprocmask"); };
  (*p_pid)=fork();
  { sigset_t sig; sigemptyset(&sig);sigaddset(&sig,SIGCHLD); sigprocmask(SIG_UNBLOCK,&sig,NULL); };
  if((*p_pid) == -1) {
    olog << "Failure forking child process." << endl;
    return false;
  };
  if((*p_pid) != 0) { /* parent */
    job_counter++;
    (*ere)=re;
    return true;
  };
  /* child */
  sched_yield(); // let parent write child's pid into list
  struct rlimit lim;
  int max_files;
  if(getrlimit(RLIMIT_NOFILE,&lim) == 0) { max_files=lim.rlim_cur; }
  else { max_files=4096; };
  /* change user */
  if(!(user.SwitchUser(su))) {
    olog << "Failed switching user" << endl; sleep(10); exit(1);
  };
  if(cred) {
    // run external plugin to acquire non-unix local credentials
    if(!cred->run(subst,subst_arg)) {
      olog << "Failed to run plugin" << endl; sleep(10); exit(1);
    };
    if(cred->result() != 0) {
      olog << "Plugin failed" << endl; sleep(10); exit(1);
    };
  };
  /* close all handles inherited from parent */
  if(max_files == RLIM_INFINITY) max_files=4096;
  for(int i=0;i<max_files;i++) { close(i); };
  int h;
  /* set up stdin,stdout and stderr */
  h=open("/dev/null",O_RDONLY); 
  if(h != 0) { if(dup2(h,0) != 0) { sleep(10); exit(1); }; close(h); };
  h=open("/dev/null",O_WRONLY);
  if(h != 1) { if(dup2(h,1) != 1) { sleep(10); exit(1); }; close(h); };
  string errlog;
  if(jobid) { 
    errlog = user.ControlDir() + "/job." + jobid + ".errors";
    h=open(errlog.c_str(),O_WRONLY | O_CREAT | O_APPEND,S_IRUSR | S_IWUSR);
    if(h==-1) { h=open("/dev/null",O_WRONLY); };
  }
  else { h=open("/dev/null",O_WRONLY); };
  if(h != 2) { if(dup2(h,2) != 2) { sleep(10); exit(1); }; close(h); };
  /* setting environment  - TODO - better environment */
  if(job_proxy) {
    setenv("GLOBUS_LOCATION",globus_loc.c_str(),1);
    unsetenv("X509_USER_KEY");
    unsetenv("X509_USER_CERT");
    unsetenv("X509_USER_PROXY");
    unsetenv("X509_RUN_AS_SERVER");
    if(jobid) {
      string proxy = user.ControlDir() + "/job." + jobid + ".proxy";
      setenv("X509_USER_PROXY",proxy.c_str(),1);
      /* for Globus 2.2 set fake cert and key, or else it takes 
         those from host in case of root user */
      setenv("X509_USER_KEY",proxy.c_str(),1);
      setenv("X509_USER_CERT",proxy.c_str(),1);
    };
  };
  execv(args[0],args);
  perror("execv");
  cerr << "Failed to start external program: " << args[0] << endl;
  sleep(10); exit(1);
}

/* restore signal handlers */
void RunParallel::deinit(void) {
  if(old_sig_chld_inited) 
    sigaction(SIGCHLD,&old_sig_chld,NULL);
  if(old_sig_hup_inited) 
    sigaction(SIGHUP,&old_sig_hup,NULL);
  if(old_sig_term_inited) 
    sigaction(SIGTERM,&old_sig_term,NULL);
}

/* setup environment to catch child's death signal */
bool RunParallel::init(void) {
  struct sigaction act;
  if(!old_sig_chld_inited) { 
    act.sa_sigaction=&sig_chld;
    sigemptyset(&(act.sa_mask)); sigaddset(&(act.sa_mask),SIGHUP);
    act.sa_flags=SA_NOCLDSTOP | SA_SIGINFO;
    if(sigaction(SIGCHLD,&act,&old_sig_chld) == -1) {
      olog << "Failed setting signal handler" << endl; return false;
    };
    old_sig_chld_inited=true; 
  };
  if(!old_sig_hup_inited) { 
    act.sa_sigaction=&sig_hup;
    sigemptyset(&(act.sa_mask)); sigaddset(&(act.sa_mask),SIGCHLD);
    act.sa_flags=SA_SIGINFO;
    if(sigaction(SIGHUP,&act,&old_sig_hup) == -1) {
      olog << "Failed setting signal handler" << endl; return false;
    };
  };
  if(!old_sig_term_inited) { 
    act.sa_sigaction=&sig_term;
    sigemptyset(&(act.sa_mask)); sigaddset(&(act.sa_mask),SIGCHLD);
    act.sa_flags=SA_SIGINFO;
    if(sigaction(SIGTERM,&act,&old_sig_hup) == -1) {
      olog << "Failed setting signal handler" << endl; return false;
    };
  };
  return true;
}

/* constructor - just init handlers */
RunParallel::RunParallel(globus_cond_t *cond_) {
  initialized=false;
  if(!init()) { deinit(); return; }
  cond=cond_;
  initialized=true;
}

/* destructor - just kill processes */
RunParallel::~RunParallel(void) {
  if(initialized) {
    for(RunElementPointer p=begin;p.is_end();p.go_next()) p->kill();
    deinit();
  };
}

bool RunParallel::was_hup(void) {
  bool res = hup_detected;
  hup_detected = false;
  return res;
};

void RunParallel::sig_term(int signum,siginfo_t *info,void* arg) {
  /* propagate to all children */
  kill(0,SIGTERM);
  exit(1);
}

void RunParallel::sig_hup(int signum,siginfo_t *info,void* arg) {
  hup_detected = true;
  if(cond) { globus_cond_signal(cond); };
//  else { olog << "Warning: sleep condition is undefined" << endl; };
  /* assuming that with SA_SIGINFO always come proper handler */
  if(old_sig_hup_inited) {
    if(old_sig_hup.sa_flags & SA_SIGINFO) {
      (*(old_sig_chld.sa_sigaction))(signum,info,arg);
    }
    else {
      if(old_sig_hup_inited) { 
#ifndef _BROKEN_SIGNAL_HANDLER_DEFINITION_
        if((old_sig_hup.sa_handler != SIG_IGN) && 
           (old_sig_hup.sa_handler != SIG_ERR) && 
           (old_sig_hup.sa_handler != SIG_DFL)) {
          (*(old_sig_chld.sa_handler))(signum);
#else
        typedef void (*signal_handler_t) (int);
        if(((signal_handler_t)(old_sig_hup.sa_handler) != SIG_IGN) && 
           ((signal_handler_t)(old_sig_hup.sa_handler) != SIG_ERR) && 
           ((signal_handler_t)(old_sig_hup.sa_handler) != SIG_DFL)) {
          (*((signal_handler_t)(old_sig_chld.sa_handler)))(signum);
#endif
        };
      };
    };
  };
}

void RunParallel::sig_chld(int signum,siginfo_t *info,void* arg) {
 /* You can not be sure if current implementation of
    posix threads does not use SIGCHLD, so lets check
    only our children */
 /* NOTE - assume no other thread is deleting jobs while signal handler
    is running. */
  if(info == NULL) {
    return;
  };
  in_signal = true;
  bool child_exited = false;
#ifndef _BROKEN_SIGNAL_HANDLER_DEFINITION_
  bool our_child_exited = false;
  pid_t pid=info->si_pid;
#else
  bool our_child_exited = true;
#endif
  RunElementPointer p=begin; p.go_next();
  /* check for failed forks */
  for(;!p.is_end();p.go_next()) {
    if(p->pid == -1)  {
//      olog << "------ sig_chld: failed fork detected" << endl;
      p.prev->next=p.cur->next;  /* remove from list */
      p.cur->released=true; /* can be removed */
      p.cur=p.prev; /* jump back */
    };
  };
  p=begin; p.go_next();
  for(;!p.is_end();p.go_next()) {
#ifndef _BROKEN_SIGNAL_HANDLER_DEFINITION_
    if(p->pid == pid) {  /* our child */
      our_child_exited = true;
    };
#endif
    if(p->pid <= 0) continue;
    /* remove zombies */ 
    /* !!!! ATTENTION - it looks like signals can be lost
       or are not queued properly (the same). So check ALL
       childern here */
    int status;  
    pid_t ch_id=waitpid(p->pid,&status,WNOHANG);
//    if(ch_id == -1) {
//      olog << "sig_chld: Warning: error in waitpid" << endl;
//    };
    if(ch_id == p->pid) {
      job_counter--;
// olog<<"***** child exited "<<job_counter<<endl;
      child_exited = true;
      if(WIFEXITED(status)) {
        p->exit_code=WEXITSTATUS(status); 
//        olog<<"Child process exited with exit code "<<status<<endl;
      }
      else {
        /* child did not exit itself - can retry ? */
        p->exit_code=2;  
//        olog<<"Child process was killed"<<endl;
      };
      p->pid=-1; /* no more running process */
      p.prev->next=p.cur->next;  /* remove from list */
      p.cur=p.prev; /* jump back */
    };
  };
  if(child_exited) {
    /* wake up main loop */
    /* not using mutex because can't be sure this won't lock forewer */
    if(cond) { globus_cond_signal(cond); };
//    else { olog << "Warning: sleep condition is undefined" << endl; };
  };
  in_signal = false;
  if(our_child_exited) return;
//  olog << "------ sig_chld: unknown child exited" << endl;
  if(old_sig_chld_inited) {
    if(old_sig_chld.sa_flags & SA_SIGINFO) {
      (*(old_sig_chld.sa_sigaction))(signum,info,arg);
    } 
    else {
      if(old_sig_chld_inited) { 
#ifndef _BROKEN_SIGNAL_HANDLER_DEFINITION_
        if((old_sig_chld.sa_handler != SIG_IGN) && 
           (old_sig_chld.sa_handler != SIG_ERR) && 
           (old_sig_chld.sa_handler != SIG_DFL)) {
          (*(old_sig_chld.sa_handler))(signum);
#else
        typedef void (*signal_handler_t) (int);
        if(((signal_handler_t)(old_sig_chld.sa_handler) != SIG_IGN) && 
           ((signal_handler_t)(old_sig_chld.sa_handler) != SIG_ERR) && 
           ((signal_handler_t)(old_sig_chld.sa_handler) != SIG_DFL)) {
          (*((signal_handler_t)(old_sig_chld.sa_handler)))(signum);
#endif
        };
      };
    };
  };
}

bool RunParallel::plain_run_piped(char *const args[],const string *Din,string *Dout,string *Derr,unsigned int timeout,int* result) {
  int p[2];
  // establsh pipes and/or handles
  int din = -1;
  int dout = -1;
  int derr = -1;
  int din_ = -1;
  int dout_ = -1;
  int derr_ = -1;
  if(Din) {
    if(pipe(p) == 0) { din=p[1]; din_=p[0]; };
  } else {
    din_=open("/dev/null",O_RDONLY);
  };
  if(Dout) {
    if(pipe(p) == 0) { dout=p[0]; dout_=p[1]; };
  } else {
    dout_=open("/dev/null",O_WRONLY);
  };
  if(Derr) {
    if(pipe(p) == 0) { derr=p[0]; derr_=p[1]; };
  } else {
    derr_=open("/dev/null",O_WRONLY);
  };
  if((din_==-1) || (dout_==-1) || (derr_==-1)) {
    olog << "Failure opening pipes." << endl;
    if(din != -1) close(din);
    if(din_ != -1) close(din_);
    if(dout != -1) close(dout);
    if(dout_ != -1) close(dout_);
    if(derr != -1) close(derr);
    if(derr_ != -1) close(derr_);
    return false;
  };
  /* create slot */
  RunElement* re = new RunElement;
  re=add_handled(re);
  pid_t* p_pid = &(re->pid);
  pid_t cpid;
  { sigset_t sig; sigemptyset(&sig);sigaddset(&sig,SIGCHLD); sigprocmask(SIG_BLOCK,&sig,NULL); perror("sigprocmask"); };
  (*p_pid)=fork();
  { sigset_t sig; sigemptyset(&sig);sigaddset(&sig,SIGCHLD); sigprocmask(SIG_UNBLOCK,&sig,NULL); };
  if((*p_pid)==-1) {
    olog << "Failure forking child process." << endl;
    release(re);
    if(din != -1) close(din);
    if(din_ != -1) close(din_);
    if(dout != -1) close(dout);
    if(dout_ != -1) close(dout_);
    if(derr != -1) close(derr);
    if(derr_ != -1) close(derr_);
    return false;
  };
  if((*p_pid) != 0) { /* parent */
    job_counter++; close(din_); close(dout_); close(derr_);
    /* pipe to stdin */
    if(Din) { write(din,Din->c_str(),Din->length()); close(din); din=-1; };
    time_t ct = time(NULL);
    time_t lt = ct + timeout;
    if(Dout || Derr) {
    /* read from stdout and stderr */
      int sdmax = din; if(dout>sdmax) sdmax=dout; if(derr>sdmax) sdmax=derr;
      fd_set sdin,sdout,sderr;
      for(;;) {
        FD_ZERO(&sdin); FD_ZERO(&sdout); FD_ZERO(&sderr);
        if(dout!=-1) FD_SET(dout,&sdin);
        if(derr!=-1) FD_SET(derr,&sdin);
        struct timeval to;
        to.tv_usec=0; to.tv_sec=lt-ct;
        int n = select(sdmax+1,&sdin,&sdout,&sderr,&to);
        if(n==0) { // timeout
          olog << "Timeout waiting for child to finish"<< endl;
	  if(dout!=-1) { close(dout); dout=-1; };
	  if(derr!=-1) { close(derr); derr=-1; };
          return false; 
        };
        if(dout!=-1) if(FD_ISSET(dout,&sdin)) {
	  char buf[256];
	  int l = read(dout,buf,255);
	  if(l==0) { close(dout); dout=-1; } // eof
	  else { buf[l]=0; (*Dout)+=buf; };
        };
        if(derr!=-1) if(FD_ISSET(derr,&sdin)) {
	  char buf[256];
	  int l = read(derr,buf,255);
	  if(l==0) { close(derr); derr=-1; } // eof
	  else { buf[l]=0; (*Derr)+=buf; };
        };
	if((dout==-1) && (derr==-1)) break;
        ct=time(NULL);
	if(ct>=lt) { // timeout
          olog << "Timeout waiting for child to finish"<< endl;
	  if(dout!=-1) { close(dout); dout=-1; };
	  if(derr!=-1) { close(derr); derr=-1; };
          return false; 
	};
      };
    };
    /* wait for job to finish */
    while(re->get_exit_code() == -1) {
      ct=time(NULL);
      if(ct>=lt) {
        olog << "Timeout waiting for child to finish"<< endl;
        return false;
      };
      usleep(100000);
    };
    if(result) (*result)=re->get_exit_code();
    return true;
  };
  /* child */
  sched_yield();
  /* set up stdin,stdout and stderr */
  close(0); close(1); close(2);
  if(dup2(din_,0) != 0) { exit(1); };
  if(dup2(dout_,1) != 1) { exit(1); };
  if(dup2(derr_,2) != 2) { exit(1); };
  struct rlimit lim;
  int max_files = 4096;
  if(getrlimit(RLIMIT_NOFILE,&lim) != 0) { max_files=lim.rlim_cur; };
  /* close all handles inherited from parent */
  if(max_files == RLIM_INFINITY) max_files=4096;
  for(int i=3;i<max_files;i++) { close(i); };
  /* run executable */
  execv(args[0],args);
  perror("execv");
  cerr<<"Failed to start external program: "<<args[0]<<endl;
  exit(1);
}

void free_args(char** args) {
  if(args == NULL) return;
  for(int i=0;args[i];i++) free(args[i]);
  free(args);
};

char** string_to_args(const string& command) {
  if(command.length() == 0) return NULL;
  int n = 100;
  char** args = (char**)malloc(n*sizeof(char**));
  int i;
  for(i=0;i<n;i++) args[i]=NULL;
  if(args == NULL) return NULL;
  string args_s = command;
  string arg_s;
  for(i=0;;i++) {
    if(i==(n-1)) {
      n+=10;
      char** args_ = (char**)realloc(args,n*sizeof(char**));
      if(args_ == NULL) {
        free(args);
        return NULL;
      };
      args=args_;
    };
    arg_s=config_next_arg(args_s);
    if(arg_s.length() == 0) break;
    args[i]=strdup(arg_s.c_str());
    if(args[i] == NULL) {
      free(args);
      return NULL;
    };
  };
  return args;
}


void RunPlugin::set(const string& cmd) {
  args_.resize(0); lib="";
  char** args = string_to_args(cmd);
  if(args == NULL) return;
  for(char** arg = args;*arg;arg++) {
    args_.push_back(string(*arg));
  };
  free(args);
  if(args_.size() == 0) return;
  string& exc = *(args_.begin());
  if(exc[0] == '/') return;
  int n = exc.find('@');
  if(n == string::npos) return;
  int p = exc.find('/');
  if((p != string::npos) && (p < n)) return;
  lib=exc.substr(n+1); exc.resize(n);
  if(lib[0] != '/') lib="./"+lib;
}

void RunPlugin::set(char const * const * args) {
  args_.resize(0); lib="";
  if(args == NULL) return;
  for(char const * const * arg = args;*arg;arg++) {
    args_.push_back(string(*arg));
  };
  if(args_.size() == 0) return;
  string& exc = *(args_.begin());
  if(exc[0] == '/') return;
  int n = exc.find('@');
  if(n == string::npos) return;
  int p = exc.find('/');
  if((p != string::npos) && (p < n)) return;
  lib=exc.substr(n+1); exc.resize(n);
  if(lib[0] != '/') lib="./"+lib;
}

bool RunPlugin::run(void) {
  if(args_.size() == 0) return true;
  char** args = (char**)malloc(sizeof(char*)*(args_.size()+1));
  if(args == NULL) return false;
  int n = 0;
  for(list<string>::iterator i = args_.begin();
              i!=args_.end();++i,++n) args[n]=(char*)(i->c_str());
  args[n]=NULL;
  if(lib.length() == 0) {
    if(!RunParallel::plain_run_piped(args,
                        &stdin_,&stdout_,&stderr_,timeout_,&result_)) {
      free(args);
      return false;
    };
  } else {
    void* lib_h = dlopen(lib.c_str(),RTLD_NOW);
    if(lib_h == NULL) { free(args); return false; };
    lib_plugin_t f = (lib_plugin_t)dlsym(lib_h,args[0]);
    if(f == NULL) { dlclose(lib_h); free(args); return false; };
    result_ = (*f)(args[1],args[2],args[3],args[4],args[5],
                   args[6],args[7],args[8],args[9],args[10],
                   args[11],args[12],args[13],args[14],args[15],
                   args[16],args[17],args[18],args[19],args[20],
                   args[21],args[22],args[23],args[24],args[25],
                   args[26],args[27],args[28],args[29],args[30],
                   args[31],args[32],args[33],args[34],args[35],
                   args[36],args[37],args[38],args[39],args[40],
                   args[41],args[42],args[43],args[44],args[45],
                   args[56],args[57],args[58],args[59],args[60],
                   args[61],args[62],args[63],args[64],args[65],
                   args[66],args[67],args[68],args[69],args[70],
                   args[71],args[72],args[73],args[74],args[75],
                   args[76],args[77],args[78],args[79],args[80],
                   args[81],args[82],args[83],args[84],args[85],
                   args[86],args[87],args[88],args[89],args[90],
                   args[91],args[92],args[93],args[94],args[95],
                   args[96],args[97],args[98],args[99],args[100]);
    dlclose(lib_h);
  };
  free(args);
  return true;
}

bool RunPlugin::run(substitute_t subst,void* arg) {
  result_=0; stdout_=""; stderr_=="";
  if(subst == NULL) return run();
  if(args_.size() == 0) return true;
  char** args = (char**)malloc(sizeof(char*)*(args_.size()+1));
  if(args == NULL) return false;
  list<string> args__;
  for(list<string>::iterator i = args_.begin();i!=args_.end();++i) {
    args__.push_back(*i);
  };
  for(list<string>::iterator i = args__.begin();i!=args__.end();++i) {
    (*subst)(*i,arg);
  };
  int n = 0;
  for(list<string>::iterator i = args__.begin();
              i!=args__.end();++i,++n) args[n]=(char*)(i->c_str());
  args[n]=NULL;
  if(lib.length() == 0) {
    if(!RunParallel::plain_run_piped(args,
                          &stdin_,&stdout_,&stderr_,timeout_,&result_)) {
      free(args);
      return false;
    };
  } else {
    void* lib_h = dlopen(lib.c_str(),RTLD_NOW);
    if(lib_h == NULL) { 
      free(args); return false;
    };
    lib_plugin_t f = (lib_plugin_t)dlsym(lib_h,args[0]);
    if(f == NULL) { 
      dlclose(lib_h); free(args); return false;
    };
    result_ = (*f)(args[1],args[2],args[3],args[4],args[5],
                   args[6],args[7],args[8],args[9],args[10],
                   args[11],args[12],args[13],args[14],args[15],
                   args[16],args[17],args[18],args[19],args[20],
                   args[21],args[22],args[23],args[24],args[25],
                   args[26],args[27],args[28],args[29],args[30],
                   args[31],args[32],args[33],args[34],args[35],
                   args[36],args[37],args[38],args[39],args[40],
                   args[41],args[42],args[43],args[44],args[45],
                   args[56],args[57],args[58],args[59],args[60],
                   args[61],args[62],args[63],args[64],args[65],
                   args[66],args[67],args[68],args[69],args[70],
                   args[71],args[72],args[73],args[74],args[75],
                   args[76],args[77],args[78],args[79],args[80],
                   args[81],args[82],args[83],args[84],args[85],
                   args[86],args[87],args[88],args[89],args[90],
                   args[91],args[92],args[93],args[94],args[95],
                   args[96],args[97],args[98],args[99],args[100]);
    dlclose(lib_h);
  };
  free(args);
  return true;
}

void RunPlugins::add(const string& cmd) {
  RunPlugin* r = new RunPlugin(cmd);
  if(!r) return;
  if(!(*r)) return;
  plugins_.push_back(r);
}

bool RunPlugins::run(void) {
  for(list<RunPlugin*>::iterator r = plugins_.begin();
                                  r != plugins_.end();++r) {
    if(!(*r)->run()) return false;
    if((*r)->result() != 0) {
      result_=result(); return true;
    };
  };
  result_=0;
  return true;
}

bool RunPlugins::run(RunPlugin::substitute_t subst,void* arg) {
  for(list<RunPlugin*>::iterator r = plugins_.begin();
                                  r != plugins_.end();++r) {
    if(!(*r)->run(subst,arg)) return false;
    if((*r)->result() != 0) {
      result_=result(); return true;
    };
  };
  result_=0;
  return true;
}

